@@ -1,3 +1,4 @@
+# pylint:disable=missing-class-docstring
 import logging
 
 import pyvex
@@ -19,7 +20,6 @@ from .expression import (
     ITE,
     Reinterpret,
     VEXCCallExpression,
-    TernaryOp,
 )
 from .converter_common import SkipConversionNotice, Converter
 
@@ -43,6 +43,9 @@ class VEXExprConverter(Converter):
         :param expr:
         :return:
         """
+        if isinstance(expr, pyvex.const.IRConst):
+            return VEXExprConverter.const_n(expr, manager)
+
         func = EXPRESSION_MAPPINGS.get(type(expr))
         if func is not None:
             # When something goes wrong, return a DirtyExpression instead of crashing the program
@@ -50,33 +53,17 @@ class VEXExprConverter(Converter):
                 return func(expr, manager)
             except UnsupportedIROpError:
                 log.warning("VEXExprConverter: Unsupported IROp %s.", expr.op)
-                return DirtyExpression(manager.next_atom(), expr, bits=expr.result_size(manager.tyenv))
-
-        if isinstance(expr, pyvex.const.IRConst):
-            return VEXExprConverter.const_n(expr, manager)
-
-        if isinstance(expr, pyvex.IRExpr.CCall):
-            operands = tuple(VEXExprConverter.convert(arg, manager) for arg in expr.args)
-            ccall = VEXCCallExpression(
-                manager.next_atom(),
-                expr.cee.name,
-                operands,
-                bits=expr.result_size(manager.tyenv),
-                ins_addr=manager.ins_addr,
-                vex_block_addr=manager.block_addr,
-                vex_stmt_idx=manager.vex_stmt_idx,
-            )
-            return DirtyExpression(
-                manager.next_atom(),
-                ccall,
-                bits=expr.result_size(manager.tyenv),
-                ins_addr=manager.ins_addr,
-                vex_block_addr=manager.block_addr,
-                vex_stmt_idx=manager.vex_stmt_idx,
-            )
+                return DirtyExpression(
+                    manager.next_atom(), f"unsupported_{expr.op}", [], bits=expr.result_size(manager.tyenv)
+                )
 
         log.warning("VEXExprConverter: Unsupported VEX expression of type %s.", type(expr))
-        return DirtyExpression(manager.next_atom(), expr, bits=expr.result_size(manager.tyenv))
+        try:
+            bits = expr.result_size(manager.tyenv)
+        except ValueError:
+            # e.g., "ValueError: Type Ity_INVALID does not have size"
+            bits = 0
+        return DirtyExpression(manager.next_atom(), f"unsupported_{str(type(expr))}", [], bits=bits)
 
     @staticmethod
     def convert_list(exprs, manager):
@@ -205,6 +192,7 @@ class VEXExprConverter(Converter):
             manager.next_atom(),
             op_name,
             VEXExprConverter.convert(expr.args[0], manager),
+            bits=expr.result_size(manager.tyenv),
             ins_addr=manager.ins_addr,
             vex_block_addr=manager.block_addr,
             vex_stmt_idx=manager.vex_stmt_idx,
@@ -223,9 +211,13 @@ class VEXExprConverter(Converter):
             operands[1] = Const(operands[1].idx, None, (1 << op1_bits) - op1_val, op1_bits)
 
         signed = False
+        vector_count = None
+        vector_size = None
         if op._vector_count is not None and op._vector_size is not None:
             # SIMD conversions
             op_name += "V"  # vectorized
+            vector_count = op._vector_count
+            vector_size = op._vector_size
         elif op_name in {"CmpLE", "CmpLT", "CmpGE", "CmpGT", "Div", "DivMod", "Mod", "Mul", "Mull"}:
             if op.is_signed:
                 signed = True
@@ -295,6 +287,69 @@ class VEXExprConverter(Converter):
 
         bits = op._output_size_bits
 
+        if op_name == "DivMod":
+            op1_size = op._from_size if op._from_size is not None else operands[0].bits
+            op2_size = op._to_size if op._to_size is not None else operands[1].bits
+
+            if op2_size < op1_size:
+                # e.g., DivModU64to32
+                operands[1] = Convert(
+                    manager.next_atom(),
+                    op2_size,
+                    op1_size,
+                    op._from_signed != "U",
+                    operands[1],
+                    ins_addr=manager.ins_addr,
+                    vex_block_addr=manager.block_addr,
+                    vex_stmt_idx=manager.vex_stmt_idx,
+                )
+            chunk_bits = bits // 2
+
+            div = BinaryOp(
+                manager.next_atom(),
+                "Div",
+                operands,
+                signed,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+                bits=op1_size,
+            )
+            truncated_div = Convert(
+                manager.next_atom(),
+                op1_size,
+                chunk_bits,
+                signed,
+                div,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+            )
+            mod = BinaryOp(
+                manager.next_atom(),
+                "Mod",
+                operands,
+                signed,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+                bits=op1_size,
+            )
+            truncated_mod = Convert(
+                manager.next_atom(),
+                op1_size,
+                chunk_bits,
+                signed,
+                mod,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+            )
+
+            operands = [truncated_mod, truncated_div]
+            op_name = "Concat"
+            signed = False
+
         return BinaryOp(
             manager.next_atom(),
             op_name,
@@ -304,6 +359,8 @@ class VEXExprConverter(Converter):
             vex_block_addr=manager.block_addr,
             vex_stmt_idx=manager.vex_stmt_idx,
             bits=bits,
+            vector_count=vector_count,
+            vector_size=vector_size,
         )
 
     @staticmethod
@@ -331,14 +388,9 @@ class VEXExprConverter(Converter):
                 bits=bits,
             )
 
-        return TernaryOp(
-            manager.next_atom(),
-            op_name,
-            operands,
-            ins_addr=manager.ins_addr,
-            vex_block_addr=manager.block_addr,
-            vex_stmt_idx=manager.vex_stmt_idx,
-            bits=bits,
+        raise TypeError(
+            "Please figure out what kind of operation this is (smart money says fused multiply) and convert it into "
+            "multiple binops"
         )
 
     @staticmethod
@@ -383,6 +435,19 @@ class VEXExprConverter(Converter):
             vex_stmt_idx=manager.vex_stmt_idx,
         )
 
+    @staticmethod
+    def CCall(expr: pyvex.IRExpr.CCall, manager):
+        operands = [VEXExprConverter.convert(arg, manager) for arg in expr.args]
+        return VEXCCallExpression(
+            manager.next_atom(),
+            expr.cee.name,
+            operands,
+            bits=expr.result_size(manager.tyenv),
+            ins_addr=manager.ins_addr,
+            vex_block_addr=manager.block_addr,
+            vex_stmt_idx=manager.vex_stmt_idx,
+        )
+
 
 EXPRESSION_MAPPINGS = {
     pyvex.IRExpr.RdTmp: VEXExprConverter.RdTmp,
@@ -395,6 +460,7 @@ EXPRESSION_MAPPINGS = {
     pyvex.const.U64: VEXExprConverter.const_n,
     pyvex.IRExpr.Load: VEXExprConverter.Load,
     pyvex.IRExpr.ITE: VEXExprConverter.ITE,
+    pyvex.IRExpr.CCall: VEXExprConverter.CCall,
 }
 
 
@@ -412,7 +478,8 @@ class VEXStmtConverter(Converter):
         try:
             func = STATEMENT_MAPPINGS[type(stmt)]
         except KeyError:
-            return DirtyStatement(idx, stmt, ins_addr=manager.ins_addr)
+            dirty = DirtyExpression(manager.next_atom(), str(stmt), [], bits=0)
+            return DirtyStatement(idx, dirty, ins_addr=manager.ins_addr)
 
         return func(idx, stmt, manager)
 
@@ -492,7 +559,7 @@ class VEXStmtConverter(Converter):
             "ILGop_16Sto32": (16, 32, True),
         }
 
-        dst = VEXExprConverter.tmp(stmt.dst, manager.tyenv.sizeof(stmt.dst) // 8, manager)
+        dst = VEXExprConverter.tmp(stmt.dst, manager.tyenv.sizeof(stmt.dst), manager)
         load_bits, convert_bits, signed = sizes[stmt.cvt]
         src = Load(
             manager.next_atom(),
@@ -545,11 +612,23 @@ class VEXStmtConverter(Converter):
             }[ty]
             dataHi = VEXExprConverter.convert(stmt.dataHi, manager)
             dataLo = VEXExprConverter.convert(stmt.dataLo, manager)
-            data = BinaryOp(idx, "Concat", (dataHi, dataLo), False)
+            data = BinaryOp(manager.next_atom(), "Concat", (dataHi, dataLo), False)
 
-            expdHi = Convert(idx, widen_from_bits, widen_to_bits, False, VEXExprConverter.convert(stmt.dataHi, manager))
-            expdLo = Convert(idx, widen_from_bits, widen_to_bits, False, VEXExprConverter.convert(stmt.dataLo, manager))
-            expd = BinaryOp(idx, "Concat", (expdHi, expdLo), False)
+            expdHi = Convert(
+                manager.next_atom(),
+                widen_from_bits,
+                widen_to_bits,
+                False,
+                VEXExprConverter.convert(stmt.dataHi, manager),
+            )
+            expdLo = Convert(
+                manager.next_atom(),
+                widen_from_bits,
+                widen_to_bits,
+                False,
+                VEXExprConverter.convert(stmt.dataLo, manager),
+            )
+            expd = BinaryOp(manager.next_atom(), "Concat", (expdHi, expdLo), False)
         else:
             narrow_to_bits = widen_to_bits = None
             data = VEXExprConverter.convert(stmt.dataLo, manager)
@@ -571,9 +650,9 @@ class VEXStmtConverter(Converter):
             size,
             stmt.endness,
         )
-        cmp = BinaryOp(idx, "CmpEQ", (val, expd), False)
+        cmp = BinaryOp(manager.next_atom(), "CmpEQ", (val, expd), False)
         store = Store(
-            idx,
+            manager.next_atom(),
             addr.copy(),
             data,
             size,
@@ -586,13 +665,14 @@ class VEXStmtConverter(Converter):
         stmts.append(store)
 
         if double:
-            val_shifted = BinaryOp(idx, "Shr", (val, narrow_to_bits), False)
-            valHi = Convert(idx, widen_to_bits, narrow_to_bits, False, val_shifted)
-            valLo = Convert(idx, widen_to_bits, narrow_to_bits, False, val)
+            narrow_to_bits_con = Const(manager.next_atom(), None, narrow_to_bits, 8)
+            val_shifted = BinaryOp(manager.next_atom(), "Shr", [val, narrow_to_bits_con], False)
+            valHi = Convert(manager.next_atom(), widen_to_bits, narrow_to_bits, False, val_shifted)
+            valLo = Convert(manager.next_atom(), widen_to_bits, narrow_to_bits, False, val)
 
             wrtmp_0 = Assignment(
-                idx,
-                Tmp(idx, None, stmt.oldLo, narrow_to_bits),
+                manager.next_atom(),
+                Tmp(manager.next_atom(), None, stmt.oldLo, narrow_to_bits),
                 valLo,
                 ins_addr=manager.ins_addr,
                 vex_block_addr=manager.block_addr,
@@ -600,7 +680,7 @@ class VEXStmtConverter(Converter):
             )
             wrtmp_1 = Assignment(
                 idx,
-                Tmp(idx, None, stmt.oldHi, narrow_to_bits),
+                Tmp(manager.next_atom(), None, stmt.oldHi, narrow_to_bits),
                 valHi,
                 ins_addr=manager.ins_addr,
                 vex_block_addr=manager.block_addr,
@@ -611,7 +691,7 @@ class VEXStmtConverter(Converter):
         else:
             wrtmp = Assignment(
                 idx,
-                Tmp(idx, None, stmt.oldLo, size),
+                Tmp(manager.next_atom(), None, stmt.oldLo, size),
                 val,
                 ins_addr=manager.ins_addr,
                 vex_block_addr=manager.block_addr,
@@ -621,6 +701,44 @@ class VEXStmtConverter(Converter):
 
         return stmts
 
+    @staticmethod
+    def Dirty(idx, stmt: pyvex.IRStmt.Dirty, manager):
+        # we translate it into tmp = DirtyExpression() if possible
+
+        operands = [VEXExprConverter.convert(op, manager) for op in stmt.args]
+        guard = VEXExprConverter.convert(stmt.guard, manager) if stmt.guard is not None else None
+        bits = manager.tyenv.sizeof(stmt.tmp) if stmt.tmp != 0xFFFFFFFF else 0
+        maddr = VEXExprConverter.convert(stmt.mAddr, manager) if stmt.mAddr is not None else None
+        dirty_expr = DirtyExpression(
+            manager.next_atom(),
+            stmt.cee.name,
+            operands,
+            guard=guard,
+            mfx=stmt.mFx,
+            maddr=maddr,
+            msize=stmt.mSize,
+            bits=bits,
+        )
+
+        if stmt.tmp == 0xFFFFFFFF:
+            return DirtyStatement(
+                idx,
+                dirty_expr,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+            )
+
+        tmp = VEXExprConverter.tmp(stmt.tmp, bits, manager)
+        return Assignment(
+            idx,
+            tmp,
+            dirty_expr,
+            ins_addr=manager.ins_addr,
+            vex_block_addr=manager.block_addr,
+            vex_stmt_idx=manager.vex_stmt_idx,
+        )
+
 
 STATEMENT_MAPPINGS = {
     pyvex.IRStmt.Put: VEXStmtConverter.Put,
@@ -630,6 +748,7 @@ STATEMENT_MAPPINGS = {
     pyvex.IRStmt.StoreG: VEXStmtConverter.StoreG,
     pyvex.IRStmt.LoadG: VEXStmtConverter.LoadG,
     pyvex.IRStmt.CAS: VEXStmtConverter.CAS,
+    pyvex.IRStmt.Dirty: VEXStmtConverter.Dirty,
 }
 
 
@@ -694,15 +813,21 @@ class VEXIRSBConverter(Converter):
                 ret_reg_offset,
                 manager.arch.bits,
                 reg_name=manager.arch.translate_register_name(ret_reg_offset, size=manager.arch.bits),
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=DEFAULT_STATEMENT,
             )
             fp_ret_reg_offset = manager.arch.fp_ret_offset
-            if fp_ret_reg_offset is not None and fp_ret_reg_offset != ret_expr:
+            if fp_ret_reg_offset is not None and fp_ret_reg_offset != ret_reg_offset:
                 fp_ret_expr = Register(
                     manager.next_atom(),
                     None,
                     fp_ret_reg_offset,
                     manager.arch.bits,
                     reg_name=manager.arch.translate_register_name(fp_ret_reg_offset, size=manager.arch.bits),
+                    ins_addr=manager.ins_addr,
+                    vex_block_addr=manager.block_addr,
+                    vex_stmt_idx=DEFAULT_STATEMENT,
                 )
             else:
                 fp_ret_expr = None
@@ -710,7 +835,8 @@ class VEXIRSBConverter(Converter):
             if irsb.jumpkind == "Ijk_Call":
                 target = VEXExprConverter.convert(irsb.next, manager)
             elif irsb.jumpkind.startswith("Ijk_Sys"):
-                target = DirtyExpression(manager.next_atom(), "syscall", manager.arch.bits)
+                # FIXME: This is a hack to make syscall work. We should have a better way to handle syscalls.
+                target = DirtyExpression(manager.next_atom(), "syscall", [], bits=manager.arch.bits)
             else:
                 raise NotImplementedError("Unsupported jumpkind")
 
@@ -726,9 +852,9 @@ class VEXIRSBConverter(Converter):
                 )
             )
         elif irsb.jumpkind == "Ijk_Boring":
-            if len(conditional_jumps) == 1:
+            if conditional_jumps:
                 # fill in the false target
-                cond_jump = conditional_jumps[0]
+                cond_jump = conditional_jumps[-1]
                 cond_jump.false_target = VEXExprConverter.convert(irsb.next, manager)
 
             else:
